@@ -30,6 +30,8 @@
 /* The root of the map entry tree */
 struct master *master_list = NULL;
 
+extern long global_negative_timeout;
+
 /* Attribute to create detached thread */
 extern pthread_attr_t thread_attr;
 
@@ -39,8 +41,28 @@ static struct map_source *
 __master_find_map_source(struct master_mapent *,
 			 const char *, const char *, int, const char **);
 
-pthread_mutex_t master_mutex = PTHREAD_MUTEX_INITIALIZER;
-pthread_mutex_t instance_mutex = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t master_mutex = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t instance_mutex = PTHREAD_MUTEX_INITIALIZER;
+
+void master_mutex_lock(void)
+{
+	int status = pthread_mutex_lock(&master_mutex);
+	if (status)
+		fatal(status);
+}
+
+void master_mutex_unlock(void)
+{
+	int status = pthread_mutex_unlock(&master_mutex);
+	if (status)
+		fatal(status);
+}
+
+void master_mutex_lock_cleanup(void *arg)
+{
+	master_mutex_unlock();
+	return;
+}
 
 int master_add_autofs_point(struct master_mapent *entry,
 		time_t timeout, unsigned logopt, unsigned ghost, int submount) 
@@ -56,6 +78,7 @@ int master_add_autofs_point(struct master_mapent *entry,
 
 	ap->state_pipe[0] = -1;
 	ap->state_pipe[1] = -1;
+	ap->logpri_fifo = -1;
 
 	ap->path = strdup(entry->path);
 	if (!ap->path) {
@@ -67,52 +90,41 @@ int master_add_autofs_point(struct master_mapent *entry,
 	ap->exp_thread = 0;
 	ap->readmap_thread = 0;
 	ap->exp_timeout = timeout;
+	/*
+	 * Program command line option overrides config.
+	 * We can't use 0 negative timeout so use default.
+	 */
+	if (global_negative_timeout <= 0)
+		ap->negative_timeout = defaults_get_negative_timeout();
+	else
+		ap->negative_timeout = global_negative_timeout;
 	ap->exp_runfreq = (timeout + CHECK_RATIO - 1) / CHECK_RATIO;
-	ap->ghost = ghost;
+	ap->flags = 0;
+	if (ghost)
+		ap->flags = MOUNT_FLAG_GHOST;
 
 	if (ap->path[1] == '-')
 		ap->type = LKP_DIRECT;
 	else
 		ap->type = LKP_INDIRECT;
 
-	ap->dir_created = 0;
 	ap->logopt = logopt;
 
 	ap->parent = NULL;
+	ap->thid = 0;
 	ap->submnt_count = 0;
 	ap->submount = submount;
 	INIT_LIST_HEAD(&ap->mounts);
 	INIT_LIST_HEAD(&ap->submounts);
-
-	status = pthread_mutex_init(&ap->state_mutex, NULL);
-	if (status) {
-		free(ap->path);
-		free(ap);
-		return 0;
-	}
+	ap->shutdown = 0;
 
 	status = pthread_mutex_init(&ap->mounts_mutex, NULL);
 	if (status) {
-		status = pthread_mutex_destroy(&ap->state_mutex);
-		if (status)
-			fatal(status);
 		free(ap->path);
 		free(ap);
 		return 0;
 	}
 
-	status = pthread_cond_init(&ap->mounts_cond, NULL);
-	if (status) {
-		status = pthread_mutex_destroy(&ap->mounts_mutex);
-		if (status)
-			fatal(status);
-		status = pthread_mutex_destroy(&ap->state_mutex);
-		if (status)
-			fatal(status);
-		free(ap->path);
-		free(ap);
-		return 0;
-	}
 	entry->ap = ap;
 
 	return 1;
@@ -125,18 +137,10 @@ void master_free_autofs_point(struct autofs_point *ap)
 	if (!ap)
 		return;
 
-	status = pthread_mutex_destroy(&ap->state_mutex);
-	if (status)
-		fatal(status);
-
 	status = pthread_mutex_destroy(&ap->mounts_mutex);
 	if (status)
 		fatal(status);
 
-	status = pthread_cond_destroy(&ap->mounts_cond);
-	if (status)
-		fatal(status);
-
 	free(ap->path);
 	free(ap);
 }
@@ -283,11 +287,9 @@ struct map_source *master_find_map_source(struct master_mapent *entry,
 {
 	struct map_source *source = NULL;
 
-	master_mutex_lock();
-
+	master_source_readlock(entry);
 	source = __master_find_map_source(entry, type, format, argc, argv);
-
-	master_mutex_unlock();
+	master_source_unlock(entry);
 
 	return source;
 }
@@ -478,7 +480,7 @@ void send_map_update_request(struct autofs_point *ap)
 	struct map_source *map;
 	int status, need_update = 0;
 
-	if (!ap->ghost)
+	if (!(ap->flags & MOUNT_FLAG_GHOST))
 		return;
 
 	status = pthread_mutex_lock(&instance_mutex);
@@ -507,13 +509,7 @@ void send_map_update_request(struct autofs_point *ap)
 	if (!need_update)
 		return;
 
-	status = pthread_mutex_lock(&ap->state_mutex);
-	if (status)
-		fatal(status);
-	nextstate(ap->state_pipe[1], ST_READMAP);
-	status = pthread_mutex_unlock(&ap->state_mutex);
-	if (status)
-		fatal(status);
+	st_add_task(ap, ST_READMAP);
 
 	return;
 }
@@ -524,8 +520,7 @@ void master_source_writelock(struct master_mapent *entry)
 
 	status = pthread_rwlock_wrlock(&entry->source_lock);
 	if (status) {
-		error(LOGOPT_ANY,
-		      "master_mapent source write lock failed");
+		logmsg("master_mapent source write lock failed");
 		fatal(status);
 	}
 	return;
@@ -537,8 +532,7 @@ void master_source_readlock(struct master_mapent *entry)
 
 	status = pthread_rwlock_rdlock(&entry->source_lock);
 	if (status) {
-		error(LOGOPT_ANY,
-		      "master_mapent source read lock failed");
+		logmsg("master_mapent source read lock failed");
 		fatal(status);
 	}
 	return;
@@ -550,8 +544,7 @@ void master_source_unlock(struct master_mapent *entry)
 
 	status = pthread_rwlock_unlock(&entry->source_lock);
 	if (status) {
-		error(LOGOPT_ANY,
-		      "master_mapent source unlock failed");
+		logmsg("master_mapent source unlock failed");
 		fatal(status);
 	}
 	return;
@@ -572,7 +565,7 @@ void master_source_current_wait(struct master_mapent *entry)
 
 	status = pthread_mutex_lock(&entry->current_mutex);
 	if (status) {
-		error(LOGOPT_ANY, "entry current source lock failed");
+		logmsg("entry current source lock failed");
 		fatal(status);
 	}
 
@@ -580,8 +573,7 @@ void master_source_current_wait(struct master_mapent *entry)
 		status = pthread_cond_wait(
 				&entry->current_cond, &entry->current_mutex);
 		if (status) {
-			error(LOGOPT_ANY,
-			      "entry current source condition wait failed");
+			logmsg("entry current source condition wait failed");
 			fatal(status);
 		}
 	}
@@ -595,14 +587,13 @@ void master_source_current_signal(struct master_mapent *entry)
 
 	status = pthread_cond_signal(&entry->current_cond);
 	if (status) {
-		error(LOGOPT_ANY,
-		      "entry current source condition signal failed");
+		logmsg("entry current source condition signal failed");
 		fatal(status);
 	}
 
 	status = pthread_mutex_unlock(&entry->current_mutex);
 	if (status) {
-		error(LOGOPT_ANY, "entry current source unlock failed");
+		logmsg("entry current source unlock failed");
 		fatal(status);
 	}
 
@@ -632,6 +623,34 @@ struct master_mapent *master_find_mapent(struct master *master, const char *path
 	return NULL;
 }
 
+struct autofs_point *__master_find_submount(struct autofs_point *ap, const char *path)
+{
+	struct list_head *head, *p;
+
+	head = &ap->submounts;
+	list_for_each(p, head) {
+		struct autofs_point *submount;
+
+		submount = list_entry(p, struct autofs_point, mounts);
+
+		if (!strcmp(submount->path, path))
+			return submount;
+	}
+
+	return NULL;
+}
+
+struct autofs_point *master_find_submount(struct autofs_point *ap, const char *path)
+{
+	struct autofs_point *submount;
+
+	mounts_mutex_lock(ap);
+	submount = __master_find_submount(ap, path);
+	mounts_mutex_unlock(ap);
+
+	return submount;
+}
+
 struct master_mapent *master_new_mapent(struct master *master, const char *path, time_t age)
 {
 	struct master_mapent *entry;
@@ -688,17 +707,13 @@ void master_remove_mapent(struct master_mapent *entry)
 	if (entry->ap->submount)
 		return;
 
-	master_mutex_lock();
 	if (!list_empty(&entry->list))
 		list_del_init(&entry->list);
-	master_mutex_unlock();
 	return;
 }
 
 void master_free_mapent_sources(struct master_mapent *entry, unsigned int free_cache)
 {
-	master_source_writelock(entry);
-
 	if (entry->maps) {
 		struct map_source *m, *n;
 
@@ -711,8 +726,6 @@ void master_free_mapent_sources(struct master_mapent *entry, unsigned int free_c
 		entry->maps = NULL;
 	}
 
-	master_source_unlock(entry);
-
 	return;
 }
 
@@ -770,6 +783,7 @@ struct master *master_new(const char *name, unsigned int timeout, unsigned int g
 	master->default_ghost = ghost;
 	master->default_timeout = timeout;
 	master->default_logging = defaults_get_logging();
+	master->logopt = master->default_logging;
 
 	INIT_LIST_HEAD(&master->mounts);
 
@@ -778,11 +792,12 @@ struct master *master_new(const char *name, unsigned int timeout, unsigned int g
 
 int master_read_master(struct master *master, time_t age, int readall)
 {
+	unsigned int logopt = master->logopt;
 	struct mapent_cache *nc;
 
 	nc = cache_init_null_cache(master);
 	if (!nc) {
-		error(LOGOPT_ANY,
+		error(logopt,
 		      "failed to init null map cache for %s", master->name);
 		return 0;
 	}
@@ -790,21 +805,13 @@ int master_read_master(struct master *master, time_t age, int readall)
 
 	master_init_scan();
 
-	if (!lookup_nss_read_master(master, age)) {
-		error(LOGOPT_ANY,
-		      "can't read master map %s", master->name);
-		return 0;
-	}
-
+	lookup_nss_read_master(master, age);
 	master_mount_mounts(master, age, readall);
 
 	master_mutex_lock();
 
-	if (list_empty(&master->mounts)) {
-		master_mutex_unlock();
-		error(LOGOPT_ANY, "no mounts in table");
-		return 0;
-	}
+	if (list_empty(&master->mounts))
+		warn(logopt, "no mounts in table");
 
 	master_mutex_unlock();
 
@@ -826,10 +833,9 @@ int master_submount_list_empty(struct autofs_point *ap)
 int master_notify_submount(struct autofs_point *ap, const char *path, enum states state)
 {
 	struct list_head *head, *p;
-	struct autofs_point *this;
-	pthread_t thid;
+	struct autofs_point *this = NULL;
 	size_t plen = strlen(path);
-	int status, ret = 1;
+	int ret = 1;
 
 	mounts_mutex_lock(ap);
 
@@ -868,33 +874,48 @@ int master_notify_submount(struct autofs_point *ap, const char *path, enum state
 
 		/* Now we have a submount to expire */
 
-		state_mutex_lock(this);
+		st_mutex_lock();
 
 		if (this->state == ST_SHUTDOWN) {
-			state_mutex_unlock(this);
+			this = NULL;
+			st_mutex_unlock();
 			break;
 		}
 
-		nextstate(this->state_pipe[1], state);
+		this->shutdown = ap->shutdown;
 
-		state_mutex_unlock(this);
+		__st_add_task(this, state);
 
-		thid = this->thid;
-		ap->mounts_signaled = MASTER_SUBMNT_WAIT;
-		while (ap->mounts_signaled == MASTER_SUBMNT_WAIT) {
-			status = pthread_cond_wait(&ap->mounts_cond, &ap->mounts_mutex);
-			if (status)
-				fatal(status);
+		st_mutex_unlock();
+		mounts_mutex_unlock(ap);
+
+		st_wait_task(this, state, 0);
+
+		/*
+		 * If our submount gets to state ST_SHUTDOWN we need to
+		 * wait until it goes away or changes to ST_READY.
+		 */
+		mounts_mutex_lock(ap);
+		st_mutex_lock();
+		while ((this = __master_find_submount(ap, path))) {
+			struct timespec t = { 0, 300000000 };
+			struct timespec r;
+
+			if (this->state != ST_SHUTDOWN)
+				break;
+
+			st_mutex_unlock();
+			mounts_mutex_unlock(ap);
+			while (nanosleep(&t, &r) == -1 && errno == EINTR)
+				memcpy(&t, &r, sizeof(struct timespec));
+			mounts_mutex_lock(ap);
+			st_mutex_lock();
 		}
+		st_mutex_unlock();
+		mounts_mutex_unlock(ap);
 
-		if (ap->mounts_signaled == MASTER_SUBMNT_JOIN) {
-			status = pthread_join(thid, NULL);
-			if (status)
-				fatal(status);
-		} else
-			ret = 0;
+		return ret;
 
-		break;
 	}
 
 	mounts_mutex_unlock(ap);
@@ -902,38 +923,13 @@ int master_notify_submount(struct autofs_point *ap, const char *path, enum state
 	return ret;
 }
 
-void master_signal_submount(struct autofs_point *ap, unsigned int join)
-{
-	int status;
-
-	if (!ap->parent || !ap->submount)
-		return;
-
-	mounts_mutex_lock(ap->parent);
-
-	ap->parent->mounts_signaled = join;
-
-	if (join == MASTER_SUBMNT_JOIN) {
-		/* We are finishing up */
-		ap->parent->submnt_count--;
-		list_del(&ap->mounts);
-	}
-
-	status = pthread_cond_signal(&ap->parent->mounts_cond);
-	if (status)
-		fatal(status);
-
-	mounts_mutex_unlock(ap->parent);
-
-	return;
-}
-
 void master_notify_state_change(struct master *master, int sig)
 {
 	struct master_mapent *entry;
 	struct autofs_point *ap;
 	struct list_head *p;
-	int state_pipe, cur_state;
+	int cur_state;
+	unsigned int logopt;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
 	master_mutex_lock();
@@ -944,20 +940,21 @@ void master_notify_state_change(struct master *master, int sig)
 		entry = list_entry(p, struct master_mapent, list);
 
 		ap = entry->ap;
+		logopt = ap->logopt;
 
-		state_mutex_lock(ap);
+		st_mutex_lock();
 
 		if (ap->state == ST_SHUTDOWN)
 			goto next;
 
-		state_pipe = ap->state_pipe[1];
-
 		switch (sig) {
 		case SIGTERM:
+		case SIGINT:
 			if (ap->state != ST_SHUTDOWN_PENDING &&
 			    ap->state != ST_SHUTDOWN_FORCE) {
 				next = ST_SHUTDOWN_PENDING;
-				nextstate(state_pipe, next);
+				ap->shutdown = 1;
+				__st_add_task(ap, next);
 			}
 			break;
 #ifdef ENABLE_FORCED_SHUTDOWN
@@ -965,23 +962,24 @@ void master_notify_state_change(struct master *master, int sig)
 			if (ap->state != ST_SHUTDOWN_FORCE &&
 			    ap->state != ST_SHUTDOWN_PENDING) {
 				next = ST_SHUTDOWN_FORCE;
-				nextstate(state_pipe, next);
+				ap->shutdown = 1;
+				__st_add_task(ap, next);
 			}
 			break;
 #endif
 		case SIGUSR1:
 			assert(ap->state == ST_READY);
 			next = ST_PRUNE;
-			nextstate(state_pipe, next);
+			__st_add_task(ap, next);
 			break;
 		}
 next:
 		if (next != ST_INVAL)
-			debug(ap->logopt,
+			debug(logopt,
 			      "sig %d switching %s from %d to %d",
 			      sig, ap->path, ap->state, next);
 
-		state_mutex_unlock(ap);
+		st_mutex_unlock();
 	}
 
 	master_mutex_unlock();
@@ -992,31 +990,34 @@ next:
 
 static int master_do_mount(struct master_mapent *entry)
 {
+	struct startup_cond suc;
 	struct autofs_point *ap;
 	pthread_t thid;
 	int status;
 
-	status = pthread_mutex_lock(&suc.mutex);
-	if (status)
-		fatal(status);
+	ap = entry->ap;
 
+	if (handle_mounts_startup_cond_init(&suc)) {
+		crit(ap->logopt,
+		     "failed to init startup cond for mount %s", entry->path);
+		return 0;
+	}
+
+	suc.ap = ap;
+	suc.root = ap->path;
 	suc.done = 0;
 	suc.status = 0;
 
-	ap = entry->ap;
-
 	debug(ap->logopt, "mounting %s", entry->path);
 
-	if (pthread_create(&thid, &thread_attr, handle_mounts, ap)) {
+	status = pthread_create(&thid, &thread_attr, handle_mounts, &suc);
+	if (status) {
 		crit(ap->logopt,
 		     "failed to create mount handler thread for %s",
 		     entry->path);
-		status = pthread_mutex_unlock(&suc.mutex);
-		if (status)
-			fatal(status);
+		handle_mounts_startup_cond_destroy(&suc);
 		return 0;
 	}
-	entry->thid = thid;
 
 	while (!suc.done) {
 		status = pthread_cond_wait(&suc.cond, &suc.mutex);
@@ -1026,52 +1027,21 @@ static int master_do_mount(struct master_mapent *entry)
 
 	if (suc.status) {
 		error(ap->logopt, "failed to startup mount");
-		status = pthread_mutex_unlock(&suc.mutex);
-		if (status)
-			fatal(status);
+		handle_mounts_startup_cond_destroy(&suc);
 		return 0;
 	}
+	entry->thid = thid;
 
-	status = pthread_mutex_unlock(&suc.mutex);
-	if (status)
-		fatal(status);
+	handle_mounts_startup_cond_destroy(&suc);
 
 	return 1;
 }
 
-static void shutdown_entry(struct master_mapent *entry)
-{
-	int state_pipe;
-	struct autofs_point *ap;
-	struct stat st;
-	int ret;
-
-	ap = entry->ap;
-
-	debug(ap->logopt, "%s", entry->path);
-
-	state_mutex_lock(ap);
-
-	state_pipe = ap->state_pipe[1];
-
-	ret = fstat(state_pipe, &st);
-	if (ret == -1)
-		goto next;
-
-	nextstate(state_pipe, ST_SHUTDOWN_PENDING);
-next:
-	state_mutex_unlock(ap);
-
-	return;
-}
-
 static void check_update_map_sources(struct master_mapent *entry, int readall)
 {
 	struct map_source *source, *last;
-	int state_pipe, map_stale = 0;
 	struct autofs_point *ap;
-	struct stat st;
-	int ret;
+	int map_stale = 0;
 
 	if (readall)
 		map_stale = 1;
@@ -1124,17 +1094,8 @@ static void check_update_map_sources(struct master_mapent *entry, int readall)
 	master_source_unlock(entry);
 
 	/* The map sources have changed */
-	if (map_stale) {
-		state_mutex_lock(ap);
-
-		state_pipe = entry->ap->state_pipe[1];
-
-		ret = fstat(state_pipe, &st);
-		if (ret != -1)
-			nextstate(state_pipe, ST_READMAP);
-
-		state_mutex_unlock(ap);
-	}
+	if (map_stale)
+		st_add_task(ap, ST_READMAP);
 
 	return;
 }
@@ -1165,7 +1126,7 @@ int master_mount_mounts(struct master *master, time_t age, int readall)
 
 		/* A master map entry has gone away */
 		if (this->age < age) {
-			shutdown_entry(this);
+			st_add_task(ap, ST_SHUTDOWN_PENDING);
 			continue;
 		}
 
@@ -1173,7 +1134,7 @@ int master_mount_mounts(struct master *master, time_t age, int readall)
 		ne = cache_lookup_distinct(nc, this->path);
 		if (ne && this->age > ne->age) {
 			cache_unlock(nc);
-			shutdown_entry(this);
+			st_add_task(ap, ST_SHUTDOWN_PENDING);
 			continue;
 		}
 		nested = cache_partial_match(nc, this->path);
@@ -1187,9 +1148,7 @@ int master_mount_mounts(struct master *master, time_t age, int readall)
 		}
 		cache_unlock(nc);
 
-		check_update_map_sources(this, readall);
-
-		state_mutex_lock(ap);
+		st_mutex_lock();
 
 		state_pipe = this->ap->state_pipe[1];
 
@@ -1197,13 +1156,16 @@ int master_mount_mounts(struct master *master, time_t age, int readall)
 		ret = fstat(state_pipe, &st);
 		save_errno = errno;
 
-		state_mutex_unlock(ap);
+		st_mutex_unlock();
 
-		if (ret == -1 && save_errno == EBADF)
+		if (!ret)
+			check_update_map_sources(this, readall);
+		else if (ret == -1 && save_errno == EBADF) {
 			if (!master_do_mount(this)) {
 				list_del_init(&this->list);
 				master_free_mapent_sources(ap->entry, 1);
 				master_free_mapent(ap->entry);
+			}
 		}
 	}
 
@@ -1225,6 +1187,11 @@ int master_list_empty(struct master *master)
 	return res;
 }
 
+inline unsigned int master_get_logopt(void)
+{
+	return master_list ? master_list->logopt : LOGOPT_NONE;
+}
+
 int master_kill(struct master *master)
 {
 	if (!list_empty(&master->mounts))
@@ -1246,6 +1213,6 @@ void dump_master(struct master *master)
 	head = &master->mounts;
 	list_for_each(p, head) {
 		struct master_mapent *this = list_entry(p, struct master_mapent, list);
-		debug(LOGOPT_ANY, "path %s", this->path);
+		logmsg("path %s", this->path);
 	}
 }
